hiragana = 'ぁあぃいぅうぇえぉおかがきぎくぐけげこごさざしじすずせぜそぞただちぢっつづてでとどなにぬ' \
           'ねのはばぱひびぴふぶぷへべぺほぼぽまみむめもゃやゅゆょよらりるれろゎわゐゑをんゔゕゖゝゞ'
katakana = 'ァアィイゥウェエォオカガキギクグケゲコゴサザシジスズセゼソゾタダチヂッツヅテデトドナニヌ' \
           'ネノハバパヒビピフブプヘベペホボポマミムメモャヤュユョヨラリルレロヮワヰヱヲンヴヵヶヽヾ'
kana = hiragana + katakana

trans_hiragana2katakana = str.maketrans(hiragana, katakana)
trans_katakana2hiragana = str.maketrans(katakana, hiragana)


def hiragana2katakana(text):
    return text.translate(trans_hiragana2katakana)

def katakana2hiragana(text):
    return text.translate(trans_katakana2hiragana)


def is_hiragana(text):
    return all(c in hiragana for c in text)

def is_katakana(text):
    return all(c in katakana for c in text)

def is_kana(text):
    return all(c in kana for c in text)


def _distribute_furigana_groups(reading, reading_h, groups):
    if len(groups) < 1:
        return []

    is_kana, text, text_h = groups[0]
    text_len = len(text)

    if is_kana:
        if reading_h.startswith(text_h):
            segs = _distribute_furigana_groups(reading[text_len:], reading_h[text_len:], groups[1:])

            if segs is not None:
                furigana = '' if reading.startswith(text) else reading[:text_len]
                segs.insert(0, (text, furigana))
                return segs

        return None

    else:
        result = None

        for i in range(len(reading), text_len-1, -1):
            segs = _distribute_furigana_groups(reading[i:], reading_h[i:], groups[1:])

            if segs is not None:
                if result is not None:
                    return None     # Ambiguous

                furigana = reading[:i]
                segs.insert(0, (text, furigana))
                result = segs

            if len(groups) == 1:
                break

        return result


def distribute_furigana(expression, reading):

    if not reading or reading == expression:
        return [(expression, '')]

    groups = []
    ik_last = None

    for c in expression:
        ik = is_kana(c)
        if ik == ik_last:
            groups[-1][1] += c
        else:
            groups.append([ik, c])
            ik_last = ik

    for g in groups:
        g.append(katakana2hiragana(g[1]))

    reading_h = katakana2hiragana(reading)

    segs = _distribute_furigana_groups(reading, reading_h, groups)

    if segs:
        return segs

    return [(expression, reading)]


def distribute_furigana_html(expression, reading):
    r = []
    for txt, ruby in distribute_furigana(expression, reading):
        if not ruby:
            r.append(txt)
        else:
            r.extend['<ruby>', txt, '<rt>', ruby, '</rt></ruby>']
    return ''.join(r)



from . import tags
from . import ja_util

from enum import Enum


# Entries: (main_text, ruby_text, following_text, accent_list, dictionary_form, learning_status, is_one_t, one_t_frequency)

def parse_migaku(text):
    lines = []

    for l in text.split('\n'):
        parts = []

        last = 0
        
        while True:
            i = l.find('[', last)
            if i < 0:
                break
            j = l.find(']', i+1)
            if j < 0:
                break
            
            k = l.rfind(' ', last, i)

            if last < k:
                parts.append( (l[last:k], '', '', [], None, 2, False, 0) )

            k = max(last, k+1)

            m = l.find(' ', j+1)
            if m < 0:
                m = len(l)

            bracket_parts = l[i+1:j].split(';')
            
            # TODO: Workaround.
            if len(bracket_parts) == 3:
                bracket_parts.insert(1, '')

            ruby_text = ''
            accent_list = []
            dictionary_form = None
            learning_status = 2     # Learned
            is_one_t = False
            one_t_frequency = 0

            bracket_parts_1 = bracket_parts[0].replace('、', ',').split(',')    # Replace to support both browser and anki syntax
            ruby_text = bracket_parts_1[0]
            if len(bracket_parts_1) >= 2:
                dictionary_form = bracket_parts_1[1]
            if len(bracket_parts) >= 2:
                accent_list = bracket_parts[1].replace('、', ',').split(',')    # Replace to support both browser and anki syntax
            if len(bracket_parts) >= 3:
                learning_status = int(bracket_parts[2])
            if len(bracket_parts) >= 4:
                one_t_info_parts = bracket_parts[3].split(',')
                is_one_t = one_t_info_parts[0] == '1'
                if len(one_t_info_parts) >= 2:
                    one_t_frequency = int(one_t_info_parts[1])

            parts.append( (l[k:i], ruby_text, l[j+1:m], accent_list, dictionary_form, learning_status, is_one_t, one_t_frequency) )

            last = m+1

        if last < len(l):
            parts.append( (l[last:len(l)], '', '', [], None, 2, False, 0) )

        lines.append(parts)

    return lines


class Mode(Enum):
    KANJI = 0,
    KANJI_READING = 1,
    READING = 2,
    KANJI_READING_UNKNOWN = 3,

    @classmethod
    def from_string(cls, key):
        associations = {
            'kanji':            cls.KANJI,
            'kanjireading':     cls.KANJI_READING,
            'reading':          cls.READING,
            'unknown':          cls.KANJI_READING_UNKNOWN,
            'furigana':         cls.KANJI_READING,
            'kana':             cls.READING,
            'furigana_unknown': cls.KANJI_READING_UNKNOWN,
        }
        return associations.get(key.lower(), cls.KANJI_READING)


def frequency_color(freq):
    if freq == 0:
        return (255, 211,   0, 0.5)
    if (freq < 1500):
        return ( 44, 173, 246, 0.5)
    if (freq < 5000):
        return ( 65, 208, 182, 0.5)
    if (freq < 15000):
        return (253, 255,  22, 0.5)
    if (freq < 30000):
        return (226, 116,  32, 0.5)
    if (freq < 60000):
        return (249,  28,  28, 0.5)
    return (203, 203, 203, 0.5)


def migaku_to_ruby(parsed_lines, mode=Mode.KANJI_READING, pitch_highlighting=True, pitch_shapes=False, unknown_underlining=True, one_t_highlighting=True, one_t_frequency_marking=True):

    coloring = {
        'h': '005CE6',  # Heiban
        'a': 'E60000',  # Atamadaka
        'n': 'E68A00',  # Nakadaka
        'o': '00802B',  # Odaka
        'k': 'AC00E6',  # Kifuku
    }

    def deco_for_accent_list(accent_list):
        if pitch_highlighting and len(accent_list) and len(accent_list[0]):
            color = coloring.get(accent_list[0][0])
            if color:
                co = '{\\c&H' + color[4:6] + color[2:4] + color[0:2] + '&}'
                cc = '{\\c}'
                return co, cc
        return '', ''

    def pitch_shapes_text(accent_list):
        ret = ''
        for a in accent_list[1:]:
            color = coloring.get(a[0])
            if color:
                ret += '{\\c&H' + color[4:6] + color[2:4] + color[0:2] + '&}⬩{\\c}'
        return ret

    # List if tag lists for each line
    ret = []

    for l in parsed_lines:
        
        # Line tags
        retl = []

        for (main_text, ruby_text, following_text, accent_list, dictionary_form, learning_status, is_one_t, one_t_frequency) in l:
            co, cc = deco_for_accent_list(accent_list)

            # Unknown/1T opening tags
            if one_t_highlighting and is_one_t:
                if not one_t_frequency_marking:
                    one_t_frequency = 0
                color = frequency_color(one_t_frequency)
                retl.append( tags.TagHighlightStart(*color) )

            if unknown_underlining and learning_status < 2:
                if learning_status == 1:
                    retl.append( tags.TagUnderlineStart(241, 187, 78) )
                else:
                    retl.append( tags.TagUnderlineStart(241, 78, 78) )

            # Tags for content
            # TODO: Remove spaces from Kanji/Furigana modes
            if mode == Mode.READING:
                # Use ruby text instead of normal text if available
                txt = co + (ruby_text if ruby_text else main_text) + following_text + cc
                retl.append( tags.TagText(txt, '') )

            elif mode == Mode.KANJI:
                # Discard ruby text
                txt = co + main_text + following_text + cc
                retl.append( tags.TagText(txt, '') )

            else:   # Mode.KANJI_READING and Mode.KANJI_READING_UNKNOWN
                if ruby_text and (mode != mode.KANJI_READING_UNKNOWN or learning_status == 0):
                    groups = ja_util.distribute_furigana(main_text, ruby_text)
                    for expression, reading in groups:
                        tag_text = co + expression + cc
                        tag_ruby_text = ''
                        if reading:
                            tag_ruby_text = co + reading + cc
                        retl.append( tags.TagText(tag_text, tag_ruby_text) )
                else:
                    tag_text = co + main_text + cc
                    retl.append( tags.TagText(tag_text, '') )

                if following_text:
                    retl.append( tags.TagText(co + following_text + cc, '') )

            # Pitch shapes
            if pitch_shapes:
                retl.append( tags.TagText(pitch_shapes_text(accent_list), '') )

            # Unknown/1T closing tags
            if unknown_underlining and learning_status < 2:
                retl.append( tags.TagUnderlineEnd )

            if one_t_highlighting and is_one_t:
                retl.append( tags.TagHighlightEnd )

            # TODO: Emit spaces if in kana mode

        # Post processing to reduce number of elements
        retl_pp = []

        for tag in retl:
            # If the last and current tags are text and have no ruby, combine them
            if len(retl_pp) and tag.isof(tags.TagText) and not tag.ruby_text:
                last_tag = retl_pp[-1]
                if last_tag.isof(tags.TagText) and not last_tag.ruby_text:
                    last_tag.text = last_tag.text + tag.text
                    continue
            retl_pp.append(tag)

        ret.append(retl_pp)

    return ret


def parse(text, mode=Mode.KANJI_READING, pitch_highlighting=True, pitch_shapes=False, unknown_underlining=True, one_t_marking=True, one_t_frequency_marking=True):
    migaku_parsed = parse_migaku(text)
    return migaku_to_ruby(migaku_parsed, mode, pitch_highlighting, pitch_shapes, unknown_underlining, one_t_marking, one_t_frequency_marking)


def args_from_strings(in_args):
    out_args = [Mode.KANJI_READING, True, False, True, True, True]

    if len(in_args) >= 1:
        out_args[0] = Mode.from_string(in_args[0])

    if len(in_args) >= 2:
        out_args[1] = in_args[1].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 3:
        out_args[2] = in_args[2].lower() in ['yes', 'y', 'true', 't', '1']

    if len(in_args) >= 4:
        out_args[3] = in_args[3].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 5:
        out_args[4] = in_args[4].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 6:
        out_args[5] = in_args[5].lower() not in ['no', 'n', 'false', 'f', '0']

    return out_args


def parser_from_string_args(in_args):
    args = args_from_strings(in_args)
    return (lambda text: parse(text, *args))



from . import tags


def parse(text):
    lines = []

    for l in text.split('\n'):
        parts = []

        last = 0
        
        while True:
            i = l.find('[', last)
            if i < 0:
                break
            j = l.find(']', i+1)
            if j < 0:
                break
            
            k = l.rfind(' ', last, i)

            if last < k:
                parts.append( tags.TagText(l[last:k], '') )

            k = max(last, k+1)

            parts.append( tags.TagText(l[k:i], l[i+1:j]) )

            last = j+1

        if last < len(l):
            parts.append( tags.TagText(l[last:len(l)], '') )

        lines.append(parts)

    return lines


def parser_from_string_args(in_args):
    return parse



from . import tags

gender_colors = {
    'm': '005CE6',
    'f': 'E60000',
    'n': '808080',
}

def is_word_char(c):
    return c.isalnum()


def frequency_color(freq):
    if freq == 0:
        return (255, 211,   0, 0.5)
    if (freq < 1500):
        return ( 44, 173, 246, 0.5)
    if (freq < 5000):
        return ( 65, 208, 182, 0.5)
    if (freq < 15000):
        return (253, 255,  22, 0.5)
    if (freq < 30000):
        return (226, 116,  32, 0.5)
    if (freq < 60000):
        return (249,  28,  28, 0.5)
    return (203, 203, 203, 0.5)


def parse(text, gender_highlighting=True, unknown_underlining=True, one_t_marking=True, one_t_frequency_marking=True):
    lines_tags = []

    for line_text in text.split('\n'):
        line_tags = []

        last = 0
        
        while True:
            bracket_open = line_text.find('[', last)
            if bracket_open < 0:
                break
            bracket_close = line_text.find(']', bracket_open+1)
            if bracket_close < 0:
                break

            word_start = bracket_open
            while True:
                if word_start <= last:
                    break
                if is_word_char(line_text[word_start-1]):
                    word_start -= 1
                else:
                    break
            
            if last < word_start:
                line_tags.append(tags.TagText(line_text[last:word_start]))

            word = line_text[word_start:bracket_open]
            bracket_text = line_text[bracket_open+1:bracket_close]
            bracket_parts = bracket_text.split(';')


            if len(bracket_parts) != 3:
                line_tags.append( tags.TagText(word, '?') )
            else:
                gender = bracket_parts[0]
                learning_status = int(bracket_parts[1])

                is_one_t = False
                one_t_frequency = 0

                one_t_info_parts = bracket_parts[2].split(',')
                is_one_t = one_t_info_parts[0] == '1'
                if len(one_t_info_parts) >= 2 and one_t_frequency_marking:
                    one_t_frequency = int(one_t_info_parts[1])

                if gender_highlighting:
                    color = gender_colors.get(gender)
                    if color:
                        co = '{\\c&H' + color[4:6] + color[2:4] + color[0:2] + '&}'
                        cc = '{\\c}'
                        word = co + word + cc

                if one_t_marking and is_one_t:
                    color = frequency_color(one_t_frequency)
                    line_tags.append( tags.TagHighlightStart(*color) )

                if unknown_underlining and learning_status < 2:
                    if learning_status == 1:
                        line_tags.append( tags.TagUnderlineStart(241, 187, 78) )
                    else:
                        line_tags.append( tags.TagUnderlineStart(241, 78, 78) )

                line_tags.append( tags.TagText(word) )

                if unknown_underlining and learning_status < 2:
                    line_tags.append( tags.TagUnderlineEnd )

                if one_t_marking and is_one_t:
                    line_tags.append( tags.TagHighlightEnd )


            last = bracket_close+1

        if last < len(line_text):
            line_tags.append( tags.TagText(line_text[last:], '') )

        lines_tags.append(line_tags)
    
    return lines_tags



def args_from_strings(in_args):
    out_args = [True, True, True, True]

    if len(in_args) >= 1:
        out_args[0] = in_args[0].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 2:
        out_args[1] = in_args[1].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 3:
        out_args[2] = in_args[2].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 4:
        out_args[3] = in_args[3].lower() not in ['no', 'n', 'false', 'f', '0']

    return out_args


def parser_from_string_args(in_args):
    args = args_from_strings(in_args)
    return (lambda text: parse(text, *args))


from .rubysubs import RubySubParser, convert_sub_file
from . import tags
from . import tag_parse_ruby
from . import tag_parse_migaku_ja
from . import tag_parse_migaku_zh
from . import tag_parse_migaku_eu
from . import tag_parse_migaku_ko



import rubysubs
from PyQt5.QtGui import QGuiApplication


def main():
    import sys
    
    if len(sys.argv) < 3:
        print('Usage: %s <source subtitle> <output subtitle> [<tag parser> [arg0] [arg1] ...]' % sys.argv[0])
        sys.exit(1)

    in_path = sys.argv[1]
    out_path = sys.argv[2]

    if len(sys.argv) >= 4:
        tag_parser_requested = sys.argv[3].lower().replace('-', '_')

        tag_parser_builders = {
            'ruby':     rubysubs.tag_parse_ruby.parser_from_string_args,
            'ja':       rubysubs.tag_parse_migaku_ja.parser_from_string_args,
            'zh':       rubysubs.tag_parse_migaku_zh.parser_from_string_args,
            'zh_hk':    rubysubs.tag_parse_migaku_zh.parser_from_string_args_HK,
            'eu':       rubysubs.tag_parse_migaku_eu.parser_from_string_args,
            'ko':       rubysubs.tag_parse_migaku_ko.parser_from_string_args,
        }

        if not tag_parser_requested in tag_parser_builders:
            print('Invalid tag parser.')
            sys.exit(1)

        tag_parser = tag_parser_builders[tag_parser_requested](sys.argv[4:])
    else:
        tag_parser = rubysubs.tag_parse_ruby.parse

    # Required for QFontMetrics
    qapp = QGuiApplication(sys.argv)

    rubysubs.convert_sub_file(in_path, out_path, tag_parser)

    sys.exit(0)


if __name__ == "__main__":
    main()


class TagBaseMeta(type):

    def __repr__(cls):
        return cls.__name__

class TagBase(metaclass=TagBaseMeta):

    @classmethod
    def isof(cls, tag_type):
        return issubclass(cls, tag_type)


class TagText(TagBase):

    def __init__(self, text='', ruby_text=''):
        self.text = text
        self.ruby_text = ruby_text

    def __repr__(self):
        return '%s(\'%s\', \'%s\')' % (self.__class__.__name__, self.text, self.ruby_text)


class TagUnderlineStart(TagBase):

    def __init__(self, r=0, g=0, b=0):
        self.r = r
        self.g = g
        self.b = b

    def __repr__(self):
        return '%s(r=%d, g=%d, b=%d)' % (self.__class__.__name__, self.r, self.g, self.b)

class TagUnderlineEnd(TagBase):
    pass


class TagHighlightStart(TagBase):

    def __init__(self, r=0, g=0, b=0, a=255):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    def __repr__(self):
        return '%s(r=%d, g=%d, b=%d, a=%d)' % (self.__class__.__name__, self.r, self.g, self.b, self.a)

class TagHighlightEnd(TagBase):
    pass



import setuptools

with open('README.md', 'r', encoding='utf-8') as f_desc:
    long_description = f_desc.read()

setuptools.setup(
    name='rubysubs',
    version='0.1.4',
    author='Bent',
    author_email='bent@mail.de',
    description='Tool/library to add ruby text to subtitles',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/RicBent/rubysubs',
    packages=['rubysubs'],
    entry_points={
        'console_scripts': [
            'rubysubs = rubysubs.__main__:main',
        ],
    },
    install_requires=[
        'cchardet',
        'pysubs2',
        'PyQt5'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
    ],
)



from . import tags
from . import tag_parse_ruby

import math
import codecs
import cchardet as chardet
import pysubs2
from PyQt5.QtGui import QFont, QFontMetrics


# Removes all curly brace tags
def strip_ass_tags(text):
    ret = ''
    in_curly = False
    for c in text:
        if c == '{':
            in_curly = True
        elif c == '}':
            in_curly = False
        elif not in_curly:
            ret += c
    return ret


class RubySubParser():

    def __init__(self, frame_width, frame_height, bottom_margin, font_name, font_size, ruby_font_size, bold, tag_parser=tag_parse_ruby.parse):

        self.tag_parser = tag_parser

        # ASS font size is actually the line height, not the cap height like regularly
        probe_size = 1000
        probe_font = QFont(font_name, probe_size)
        probe_font.setBold(bold)
        probe_font_metrics = QFontMetrics(probe_font)
        ass_font_factor = probe_size / probe_font_metrics.height()

        self.normal_line_height = round(font_size)
        self.font_size = math.ceil(font_size * ass_font_factor)
        
        if ruby_font_size is None:
            self.ruby_font_size = round(font_size / 2)
        else:
            self.ruby_font_size = round(ruby_font_size * ass_font_factor)

        font = QFont(font_name, self.font_size)

        font.setBold(bold)
        self.font_metrics = QFontMetrics(font)
        self.font_height = self.font_metrics.height()

        ruby_font = QFont(font_name, self.ruby_font_size)
        ruby_font.setBold(bold)
        self.ruby_font_metrics = QFontMetrics(ruby_font)
        self.ruby_font_height = self.ruby_font_metrics.height()

        self.frame_width = frame_width
        self.sub_origin = math.floor(frame_height - (font_size/2 + bottom_margin))

    def get_line_height(self, line):
        for tag in line:
            if tag.isof(tags.TagText) and tag.ruby_text:
                return math.floor(self.normal_line_height * 1.5)
        return self.normal_line_height

    # Returns list of (layer, style, text) and postion of the next subtitle if a collision occurs
    def parse_sub(self, sub_text, start_position=None):
        sub_text = sub_text.replace('\\N', '\n')

        parsed_lines = self.tag_parser(sub_text)

        # Calculate line y positions
        # The origin of y positions is the center of the main line
        line_y_positions = [self.sub_origin if start_position is None else start_position]
        for line in reversed(parsed_lines):
            line_y_positions.insert(0, line_y_positions[0] - self.get_line_height(line))

        collision_position = line_y_positions.pop(0)

        # Process lines
        ret = []

        for line, y in zip(parsed_lines, line_y_positions):

            # Calculate part and line width
            widths = []

            for tag in line:
                if tag.isof(tags.TagText):
                    txt_no_tags = strip_ass_tags(tag.text)
                    ruby_txt_no_tags = strip_ass_tags(tag.ruby_text)

                    txt_width = self.font_metrics.horizontalAdvance(txt_no_tags)
                    ruby_txt_width = self.ruby_font_metrics.horizontalAdvance(ruby_txt_no_tags)

                    widths.append(max(txt_width, ruby_txt_width))
                else:
                    widths.append(0)

            line_width = sum(widths)

            # Center vertical position of ruby text
            ruby_y = math.floor(y - self.font_height/2 - self.ruby_font_size/2)

            # Left side of current part
            curr_x = round(self.frame_width/2 - line_width/2)

            # States for opened underlines
            underline_start_tag = None
            underline_start_x = None

            # States for opened highlights
            highlight_start_tag = None
            highlight_start_x = None
            highlight_height = None


            for (tag, width) in zip(line, widths):
                # Center of current part
                x = round(curr_x + width/2)

                # Text
                if tag.isof(tags.TagText):
                    if tag.text.strip():
                        # Insert zero width spaces around exposed spaces to prevent them collapsing
                        text = tag.text
                        text_stripped = strip_ass_tags(text)
                        if len(text_stripped) > 1:
                            if text_stripped.startswith(' '):
                                text = '\u200B' + text
                            if text_stripped.endswith(' '):
                                text = text + '\u200B'
                        ret.append( (3, 'Default', '{\\pos(%d,%d)}%s' % (x, y, text)) )

                    if tag.ruby_text.strip():
                        ret.append( (2, 'Ruby', '{\\pos(%d,%d)}%s' % (x, ruby_y, tag.ruby_text)) )
                        highlight_height = self.normal_line_height * 1.5

                # Underlines
                elif tag.isof(tags.TagUnderlineStart):
                    underline_start_tag = tag
                    underline_start_x = x

                elif tag.isof(tags.TagUnderlineEnd):
                    if underline_start_tag:
                        x1 = underline_start_x
                        x2 = x
                        y1 = math.ceil(y + self.font_height/2)
                        y2 = math.ceil(y1 + self.font_height*0.05)
                        c = '%02X%02X%02X' % (underline_start_tag.b, underline_start_tag.g, underline_start_tag.r)
                        ret.append( (1, 'Underline', '{\\pos(0,0)}{\c&H%s&}{\p1}m %d %d l %d %d %d %d %d %d{\p0}{\c}' % (c, x1, y1, x2, y1, x2, y2, x1, y2)) )

                # Highlights
                elif tag.isof(tags.TagHighlightStart):
                    highlight_start_tag = tag
                    highlight_start_x = x
                    highlight_height = self.font_height

                elif tag.isof(tags.TagHighlightEnd):
                    if highlight_start_tag:
                        x1 = highlight_start_x
                        x2 = x
                        y1 = math.ceil(y + self.font_height/2)
                        y2 = math.ceil(y1 - highlight_height)
                        c = '%02X%02X%02X' % (highlight_start_tag.b, highlight_start_tag.g, highlight_start_tag.r)
                        ca_raw = max(0, min(1, highlight_start_tag.a))
                        ca = round((1-ca_raw) * 255)
                        ret.append( (0, 'Highlight', '{\\pos(0,0)}{\c&H%s&\\1a&H%02X&}{\p1}m %d %d l %d %d %d %d %d %d{\p0}{\c\\1a}' % (c, ca, x1, y1, x2, y1, x2, y2, x1, y2)) )

                curr_x += width

        return ret, collision_position


def convert_sub_file(in_path, out_path, tag_parser=tag_parse_ruby.parse):

    # Determine encoding
    f = open(in_path, 'rb')
    subs_data = f.read()
    f.close()

    boms_for_enc = [
        ('utf-32',      (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE)),
        ('utf-16',      (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE)),
        ('utf-8-sig',   (codecs.BOM_UTF8,)),
    ]

    for enc, boms in boms_for_enc:
        if any(subs_data.startswith(bom) for bom in boms):
            subs_encoding = enc
            break
    else:
        chardet_ret = chardet.detect(subs_data)
        subs_encoding = chardet_ret['encoding']

    # Load subs
    with open(in_path, encoding=subs_encoding, errors='replace') as f:
        subs = pysubs2.SSAFile.from_file(f)

    # Determine frame size
    frame_width = int(subs.info.get('PlayResX', '1920'))
    frame_height = int(subs.info.get('PlayResY', '1080'))

    bottom_margin = round(frame_height * 0.04)

    # Create styles
    style = subs.styles.get('Default')
    if style is None:
        # TODO: Set better default style if Default is not found
        style = pysubs2.SSAStyle()
    else:
        bottom_margin = round(style.marginv)

    style.fontsize = math.floor(style.fontsize)
    style.italic = False
    style.underline = False
    style.strikeout = False
    style.scalex = 100.0
    style.scaley = 100.0
    style.spacing = 0.0
    style.angle = 0.0
    style.alignment = 5     # Center alignment
    style.marginl = 0
    style.marginr = 0
    style.marginv = 0

    ruby_style = style.copy()
    ruby_style.fontsize = math.floor(style.fontsize / 2)

    underline_style = style.copy()
    underline_style.alignment = 7   # Top-left alignment

    highlight_style = style.copy()
    highlight_style.outline = 0
    highlight_style.shadow = 0
    highlight_style.alignment = 7   # Top-left alignment

    subs.styles.clear()
    subs.styles['Default'] = style
    subs.styles['Ruby'] = ruby_style
    subs.styles['Underline'] = underline_style
    subs.styles['Highlight'] = highlight_style

    font_name = style.fontname
    font_size = style.fontsize
    ruby_font_size = ruby_style.fontsize
    bold = style.bold
    
    parser = RubySubParser(frame_width, frame_height, bottom_margin, font_name, font_size, ruby_font_size, bold, tag_parser)

    # Create parsed subtitle events
    original_events = subs.events.copy()
    original_events.sort()
    subs.events.clear()

    collision_stack = []    # (end, position)

    for e in original_events:
        position = None
        for collision_end, collision_position in reversed(collision_stack):
            if collision_end <= e.start:
                del collision_stack[-1]
            else:
                position = collision_position
                break
        
        parts, collision_position = parser.parse_sub(e.text, position)
        collision_stack.append((e.end, collision_position))

        for (layer, style, text) in parts:
            e = e.copy()
            e.layer = layer
            e.style = style
            e.text = text
            subs.events.append(e)

    subs.save(out_path, header_notice='Generated by rubysubs')



from . import tags

from enum import Enum
from itertools import zip_longest


def pinyin_number_to_tone(syllable):
    replacements = {
        'a': ['ā', 'á', 'ǎ', 'à'],
        'e': ['ē', 'é', 'ě', 'è'],
        'u': ['ū', 'ú', 'ǔ', 'ù'],
        'i': ['ī', 'í', 'ǐ', 'ì'],
        'o': ['ō', 'ó', 'ǒ', 'ò'],
        'ü': ['ǖ', 'ǘ', 'ǚ', 'ǜ'],
    }

    medials = ['i', 'u', 'ü']

    if len(syllable) < 1:
        return syllable
    
    try:
        tone_idx = int(syllable[-1])
    except ValueError:
        return syllable

    if not (1 <= tone_idx <= 5):
        return syllable

    ret = syllable.replace('v', 'ü')

    if tone_idx == 5:
        return ret[:-1]

    for i in range(len(ret)):
        c1 = ret[i]
        c2 = ret[i+1]   # safe because of tone idx

        if c1 in medials and c2 in replacements:
            return ret[:i+1] + replacements[c2][tone_idx-1] + ret[i+2:-1]
        if c1 in replacements:
            return ret[:i] + replacements[c1][tone_idx-1] + ret[i+1:-1]
    
    return syllable


def zhuyin_number_to_tone(syllable):
    tones = [ '', 'ˊ', 'ˇ', 'ˋ', '˙' ]

    if len(syllable) < 1:
        return syllable

    try:
        tone_idx = int(syllable[-1])
    except ValueError:
        return syllable

    if not (1 <= tone_idx <= 5):
        return syllable

    return syllable[:-1] + tones[tone_idx-1]



hanzi_ranges = [
    (    0x4e00,     0x9fff),
    (    0x3400,     0x4DBF),
    (0x00020000, 0x0002A6DF),
    (0x0002B740, 0x0002B81F),
    (0x0002B820, 0x0002CEAF),
    (0x0002CEB0, 0x0002EBEF),
    (    0xF900,     0xFAFF),
    (0x0002F800, 0x0002FA1F),
]

def is_hanzi(text):
    for c in text:
        return any([s <= ord(c) <= e for (s,e) in hanzi_ranges])



color_table_mandarin = [
    'E60000',
    'E68A00',
    '00802B',
    '005CE6',
    '808080',
]

color_table_cantonese = [
    'E60000',
    'E68A00',
    '00802B',
    '005CE6',
    'AC00E6',
    '808080',
]


class DecoMode(Enum):
    NONE = 0,
    PINYIN = 1,
    ZHUYIN = 2,
    JYUTPING = 3,


class Mode:

    def __init__(self, is_cantonese, mode_txt=''):
        txt = mode_txt.lower()

        self.deco_mode = DecoMode.NONE
        self.is_cantonese = is_cantonese

        if is_cantonese:
            self.only_unknown = txt.startswith('unknown')
            if self.only_unknown or txt == 'jyutping' or txt == '':
                self.deco_mode = DecoMode.JYUTPING
        else:
            self.only_unknown = txt.startswith('unknown')
            if self.only_unknown:
                txt = txt[len('unknown'):]
            if txt == 'pinyin' or txt == '':
                self.deco_mode = DecoMode.PINYIN
            elif txt == 'zhuyin':
                self.deco_mode = DecoMode.ZHUYIN


    def color_tags_for_syllable(self, syllable, tone_highlighting):
        if not tone_highlighting or len(syllable) < 1:
            return '', ''

        try:
            idx = int(syllable[-1]) - 1
        except ValueError:
            return '', ''

        color_table = color_table_cantonese if self.is_cantonese else color_table_mandarin
        idx = min(idx, len(color_table)-1)
        color = color_table[idx]

        co = '{\\c&H' + color[4:6] + color[2:4] + color[0:2] + '&}'
        cc = '{\\c}'

        return co, cc


    def markup_word(self, text, syllables, tone_highlighting, is_unknown):
        if len(text) != len(syllables):
            return (text, '?')

        ret_text = ''
        ret_deco = ''

        for h, s in zip(text, syllables):
            co, cc = self.color_tags_for_syllable(s, tone_highlighting)
            ret_text += co + h + cc
            if self.deco_mode != DecoMode.NONE and ((not self.only_unknown) or is_unknown):
                if self.deco_mode == DecoMode.PINYIN:
                    s = pinyin_number_to_tone(s)
                elif self.deco_mode == DecoMode.ZHUYIN:
                    s = zhuyin_number_to_tone(s)
                ret_deco += co + s + cc

        return (ret_text, ret_deco)


def frequency_color(freq):
    if freq == 0:
        return (255, 211,   0, 0.5)
    if (freq < 1500):
        return ( 44, 173, 246, 0.5)
    if (freq < 5000):
        return ( 65, 208, 182, 0.5)
    if (freq < 15000):
        return (253, 255,  22, 0.5)
    if (freq < 30000):
        return (226, 116,  32, 0.5)
    if (freq < 60000):
        return (249,  28,  28, 0.5)
    return (203, 203, 203, 0.5)


def parse(text, mode, tone_highlighting, unknown_underlining, one_t_marking, one_t_frequency_marking):

    lines_tags = []

    for line_text in text.split('\n'):
        line_tags = []

        last = 0

        while True:
            bracket_open = line_text.find('[', last)
            if bracket_open < 0:
                break
            bracket_close = line_text.find(']', bracket_open+1)
            if bracket_close < 0:
                break

            # Walk back and find hanzi range to decorate
            hanzi_start = bracket_open
            while True:
                if hanzi_start <= last:
                    break
                if is_hanzi(line_text[hanzi_start-1]):
                    hanzi_start -= 1
                else:
                    break
            hanzi_end = bracket_open

            if last < hanzi_start:
                line_tags.append(tags.TagText(line_text[last:hanzi_start]))

            if hanzi_start < hanzi_end:
                hanzi_text = line_text[hanzi_start:hanzi_end]
                bracket_text = line_text[bracket_open+1:bracket_close]
                bracket_parts = bracket_text.split(';')

                syllables = []
                if len(bracket_parts) >= 1:
                    syllables = bracket_parts[0].split()

                learning_status = 2
                if len(bracket_parts) >= 2:
                    learning_status = int(bracket_parts[1])
                
                is_one_t = False
                one_t_frequency = 0
                if len(bracket_parts) >= 3:
                    one_t_info_parts = bracket_parts[2].split(',')
                    is_one_t = one_t_info_parts[0] == '1'
                    if len(one_t_info_parts) >= 2 and one_t_frequency_marking:
                        one_t_frequency = int(one_t_info_parts[1])

                # Unknown/1T opening tags
                if one_t_marking and is_one_t:
                    color = frequency_color(one_t_frequency)
                    line_tags.append( tags.TagHighlightStart(*color) )

                if unknown_underlining and learning_status < 2:
                    if learning_status == 1:
                        line_tags.append( tags.TagUnderlineStart(241, 187, 78) )
                    else:
                        line_tags.append( tags.TagUnderlineStart(241, 78, 78) )

                # Hanzi/Deco
                tag_hanzi, tag_deco = mode.markup_word(hanzi_text, syllables, tone_highlighting, learning_status==0)
                line_tags.append(tags.TagText(tag_hanzi, tag_deco))                                

                # Unknown/1T closing tags
                if unknown_underlining and learning_status < 2:
                    line_tags.append( tags.TagUnderlineEnd )

                if one_t_marking and is_one_t:
                    line_tags.append( tags.TagHighlightEnd )

            last = bracket_close+1

        if last < len(line_text):
            line_tags.append(tags.TagText(line_text[last:]))

        lines_tags.append(line_tags)

    return lines_tags



def args_from_strings(in_args, is_cantonese):
    out_args = [Mode(is_cantonese), True, True, True, True]

    if len(in_args) >= 1:
        out_args[0] = Mode(is_cantonese, in_args[0])

    if len(in_args) >= 2:
        out_args[1] = in_args[1].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 3:
        out_args[2] = in_args[2].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 4:
        out_args[3] = in_args[3].lower() not in ['no', 'n', 'false', 'f', '0']

    if len(in_args) >= 5:
        out_args[4] = in_args[4].lower() not in ['no', 'n', 'false', 'f', '0']

    return out_args


def parser_from_string_args(in_args):
    args = args_from_strings(in_args, False)
    return (lambda text: parse(text, *args))

def parser_from_string_args_HK(in_args):
    args = args_from_strings(in_args, True)
    return (lambda text: parse(text, *args))
